"""
Repackaging of Google's Diff Match and Patch libraries.

Offers robust algorithms to perform the operations required for synchronizing plain text.
"""

from .__version__ import __version__
from .diff_match_patch import __author__, __doc__, diff_match_patch, patch_obj

__packager__ = "Amethyst Reese (amy@noswap.com)"
